import numpy as np

from io import StringIO
from numba import cuda, float32, float64, int32, intp
from numba.cuda.testing import unittest, CUDATestCase
from numba.cuda.testing import (
    skip_on_cudasim,
    skip_with_nvdisasm,
    skip_without_nvdisasm,
)


@skip_on_cudasim("Simulator does not generate code to be inspected")
class TestInspect(CUDATestCase):
    @property
    def cc(self):
        return cuda.current_context().device.compute_capability

    def test_monotyped(self):
        sig = (float32, int32)

        @cuda.jit(sig)
        def foo(x, y):
            pass

        file = StringIO()
        foo.inspect_types(file=file)
        typeanno = file.getvalue()
        # Function name in annotation
        self.assertIn("foo", typeanno)
        # Signature in annotation
        self.assertIn("(float32, int32)", typeanno)
        file.close()
        # Function name in LLVM
        llvm = foo.inspect_llvm(sig)
        self.assertIn("foo", llvm)

        # Kernel in LLVM
        self.assertIn("define void @", llvm)

        asm = foo.inspect_asm(sig)

        # Function name in PTX
        self.assertIn("foo", asm)
        # NVVM inserted comments in PTX
        self.assertIn("Generated by NVIDIA NVVM Compiler", asm)

    def test_polytyped(self):
        @cuda.jit
        def foo(x, y):
            pass

        foo[1, 1](1, 1)
        foo[1, 1](1.2, 2.4)

        file = StringIO()
        foo.inspect_types(file=file)
        typeanno = file.getvalue()
        file.close()
        # Signature in annotation
        self.assertIn("({0}, {0})".format(intp), typeanno)
        self.assertIn("(float64, float64)", typeanno)

        # Signature in LLVM dict
        llvmirs = foo.inspect_llvm()
        self.assertEqual(
            2,
            len(llvmirs),
        )
        self.assertIn((intp, intp), llvmirs)
        self.assertIn((float64, float64), llvmirs)

        # Function name in LLVM
        self.assertIn("foo", llvmirs[intp, intp])
        self.assertIn("foo", llvmirs[float64, float64])

        # Kernels in LLVM
        self.assertIn("define void @", llvmirs[intp, intp])
        self.assertIn("define void @", llvmirs[float64, float64])

        asmdict = foo.inspect_asm()

        # Signature in assembly dict
        self.assertEqual(
            2,
            len(asmdict),
        )
        self.assertIn((intp, intp), asmdict)
        self.assertIn((float64, float64), asmdict)

        # NVVM inserted in PTX
        self.assertIn("foo", asmdict[intp, intp])
        self.assertIn("foo", asmdict[float64, float64])

    def _test_inspect_sass(self, kernel, name, sass):
        # Ensure function appears in output
        seen_function = False
        for line in sass.split():
            if ".text" in line and name in line:
                seen_function = True
        self.assertTrue(seen_function)

        self.assertRegex(sass, r'//## File ".*/test_inspect.py", line [0-9]')

        # Some instructions common to all supported architectures that should
        # appear in the output
        self.assertIn("S2R", sass)  # Special register to register
        self.assertIn("BRA", sass)  # Branch
        self.assertIn("EXIT", sass)  # Exit program

    @skip_without_nvdisasm("nvdisasm needed for inspect_sass()")
    def test_inspect_sass_eager(self):
        sig = (float32[::1], int32[::1])

        @cuda.jit(sig, lineinfo=True)
        def add(x, y):
            i = cuda.grid(1)
            if i < len(x):
                x[i] += y[i]

        self._test_inspect_sass(add, "add", add.inspect_sass(sig))

    @skip_without_nvdisasm("nvdisasm needed for inspect_sass()")
    def test_inspect_sass_lazy(self):
        @cuda.jit(lineinfo=True)
        def add(x, y):
            i = cuda.grid(1)
            if i < len(x):
                x[i] += y[i]

        x = np.arange(10).astype(np.int32)
        y = np.arange(10).astype(np.float32)
        add[1, 10](x, y)

        signature = (int32[::1], float32[::1])
        self._test_inspect_sass(add, "add", add.inspect_sass(signature))

    @skip_with_nvdisasm(
        "Missing nvdisasm exception only generated when it is not present"
    )
    def test_inspect_sass_nvdisasm_missing(self):
        @cuda.jit((float32[::1],))
        def f(x):
            x[0] = 0

        with self.assertRaises(RuntimeError) as raises:
            f.inspect_sass()

        self.assertIn("nvdisasm has not been found", str(raises.exception))

    @skip_without_nvdisasm("nvdisasm needed for inspect_sass_cfg()")
    def test_inspect_sass_cfg(self):
        sig = (float32[::1], int32[::1])

        @cuda.jit(sig)
        def add(x, y):
            i = cuda.grid(1)
            if i < len(x):
                x[i] += y[i]

        self.assertRegex(
            add.inspect_sass_cfg(signature=sig), r"digraph\s*\w\s*{(.|\n)*\n}"
        )


if __name__ == "__main__":
    unittest.main()



import numpy as np
from numba import cuda
from numba.cuda.testing import unittest, CUDATestCase


class TestCudaEvent(CUDATestCase):
    def test_event_elapsed(self):
        N = 32
        dary = cuda.device_array(N, dtype=np.double)
        evtstart = cuda.event()
        evtend = cuda.event()

        evtstart.record()
        cuda.to_device(np.arange(N, dtype=np.double), to=dary)
        evtend.record()
        evtend.wait()
        evtend.synchronize()
        # Exercise the code path
        evtstart.elapsed_time(evtend)

    def test_event_elapsed_stream(self):
        N = 32
        stream = cuda.stream()
        dary = cuda.device_array(N, dtype=np.double)
        evtstart = cuda.event()
        evtend = cuda.event()

        evtstart.record(stream=stream)
        cuda.to_device(np.arange(N, dtype=np.double), to=dary, stream=stream)
        evtend.record(stream=stream)
        evtend.wait(stream=stream)
        evtend.synchronize()
        # Exercise the code path
        evtstart.elapsed_time(evtend)


if __name__ == "__main__":
    unittest.main()



import unittest

from numba.cuda.testing import (
    CUDATestCase,
    skip_if_cudadevrt_missing,
    skip_on_cudasim,
    skip_unless_cc_60,
    skip_if_mvc_enabled,
)
from numba.tests.support import captured_stdout


@skip_if_cudadevrt_missing
@skip_unless_cc_60
@skip_if_mvc_enabled("CG not supported with MVC")
@skip_on_cudasim("cudasim doesn't support cuda import at non-top-level")
class TestLaplace(CUDATestCase):
    """
    Test simple vector addition
    """

    def setUp(self):
        # Prevent output from this test showing up when running the test suite
        self._captured_stdout = captured_stdout()
        self._captured_stdout.__enter__()
        super().setUp()

    def tearDown(self):
        # No exception type, value, or traceback
        self._captured_stdout.__exit__(None, None, None)
        super().tearDown()

    def test_ex_laplace(self):
        # set True to regenerate the figures that
        # accompany this example
        plot = False

        # ex_laplace.import.begin
        import numpy as np
        from numba import cuda
        # ex_laplace.import.end

        # ex_laplace.allocate.begin
        # Use an odd problem size.
        # This is so there can be an element truly in the "middle" for symmetry.
        size = 1001
        data = np.zeros(size)

        # Middle element is made very hot
        data[500] = 10000
        buf_0 = cuda.to_device(data)

        # This extra array is used for synchronization purposes
        buf_1 = cuda.device_array_like(buf_0)

        niter = 10000
        # ex_laplace.allocate.end

        if plot:
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots(figsize=(16 * 0.66, 9 * 0.66))
            plt.plot(
                np.arange(len(buf_0)),
                buf_0.copy_to_host(),
                lw=3,
                marker="*",
                color="black",
            )

            plt.title("Initial State", fontsize=24)
            plt.xlabel("Position", fontsize=24)
            plt.ylabel("Temperature", fontsize=24)

            ax.set_xticks(ax.get_xticks(), fontsize=16)
            ax.set_yticks(ax.get_yticks(), fontsize=16)
            plt.xlim(0, len(data))
            plt.ylim(0, 10001)
            plt.savefig("laplace_initial.svg")

        # ex_laplace.kernel.begin
        @cuda.jit
        def solve_heat_equation(buf_0, buf_1, timesteps, k):
            i = cuda.grid(1)

            # Don't continue if our index is outside the domain
            if i >= len(buf_0):
                return

            # Prepare to do a grid-wide synchronization later
            grid = cuda.cg.this_grid()

            for step in range(timesteps):
                # Select the buffer from the previous timestep
                if (step % 2) == 0:
                    data = buf_0
                    next_data = buf_1
                else:
                    data = buf_1
                    next_data = buf_0

                # Get the current temperature associated with this point
                curr_temp = data[i]

                # Apply formula from finite difference equation
                if i == 0:
                    # Left wall is held at T = 0
                    next_temp = curr_temp + k * (data[i + 1] - (2 * curr_temp))
                elif i == len(data) - 1:
                    # Right wall is held at T = 0
                    next_temp = curr_temp + k * (data[i - 1] - (2 * curr_temp))
                else:
                    # Interior points are a weighted average of their neighbors
                    next_temp = curr_temp + k * (
                        data[i - 1] - (2 * curr_temp) + data[i + 1]
                    )

                # Write new value to the next buffer
                next_data[i] = next_temp

                # Wait for every thread to write before moving on
                grid.sync()

        # ex_laplace.kernel.end

        # ex_laplace.launch.begin
        solve_heat_equation.forall(len(data))(buf_0, buf_1, niter, 0.25)
        # ex_laplace.launch.end

        results = buf_1.copy_to_host()
        if plot:
            fig, ax = plt.subplots(figsize=(16 * 0.66, 9 * 0.66))
            plt.plot(
                np.arange(len(results)),
                results,
                lw=3,
                marker="*",
                color="black",
            )
            plt.title(f"T = {niter}", fontsize=24)
            plt.xlabel("Position", fontsize=24)
            plt.ylabel("Temperature", fontsize=24)

            ax.set_xticks(ax.get_xticks(), fontsize=16)
            ax.set_yticks(ax.get_yticks(), fontsize=16)

            plt.ylim(0, max(results))
            plt.xlim(0, len(results))
            plt.savefig("laplace_final.svg")

        # Integral over the domain should be equal to its initial value.
        # Note that this should match the initial value of data[500] above, but
        # we don't assign it to a variable because that would make the example
        # code look a bit oddly verbose.
        np.testing.assert_allclose(results.sum(), 10000)


if __name__ == "__main__":
    unittest.main()



import numpy as np

from collections import namedtuple
from itertools import product
from numba import vectorize
from numba import cuda, int32, float32, float64
from numba.cuda.cudadrv.driver import CudaAPIError, driver
from numba.cuda.testing import skip_on_cudasim
from numba.cuda.testing import CUDATestCase
import unittest


# Signatures to test with - these are all homogeneous in dtype, so the output
# dtype should match the input dtype - the output should not have been cast
# upwards, as reported in #8400: https://github.com/numba/numba/issues/8400
signatures = [
    int32(int32, int32),
    float32(float32, float32),
    float64(float64, float64),
]

# The order here is chosen such that each subsequent dtype might have been
# casted to a previously-used dtype. This is unlikely to be an issue for CUDA,
# but there might be future circumstances in which it becomes relevant, perhaps
# if it supported Dynamic UFuncs, and we want to ensure that an implementation
# for a the given dtype is used rather than casting the input upwards.
dtypes = (np.float64, np.float32, np.int32)

# NumPy ndarray orders
orders = ("C", "F")

# Input sizes corresponding to operations:
# - Less than one warp,
# - Less than one block,
# - Greater than one block (i.e. many blocks)
input_sizes = (8, 100, 2**10 + 1)


@skip_on_cudasim("ufunc API unsupported in the simulator")
class TestCUDAVectorize(CUDATestCase):
    # Presumably chosen as an odd number unlikely to coincide with the total
    # thread count, and large enough to ensure a significant number of blocks
    # are used.
    N = 1000001

    def test_scalar(self):
        @vectorize(signatures, target="cuda")
        def vector_add(a, b):
            return a + b

        a = 1.2
        b = 2.3
        c = vector_add(a, b)
        self.assertEqual(c, a + b)

    def test_1d(self):
        @vectorize(signatures, target="cuda")
        def vector_add(a, b):
            return a + b

        for ty in dtypes:
            data = np.array(np.random.random(self.N), dtype=ty)
            expected = np.add(data, data)
            actual = vector_add(data, data)
            np.testing.assert_allclose(expected, actual)
            self.assertEqual(actual.dtype, ty)

    def test_1d_async(self):
        @vectorize(signatures, target="cuda")
        def vector_add(a, b):
            return a + b

        stream = cuda.stream()

        for ty in dtypes:
            data = np.array(np.random.random(self.N), dtype=ty)
            device_data = cuda.to_device(data, stream)

            dresult = vector_add(device_data, device_data, stream=stream)
            actual = dresult.copy_to_host()

            expected = np.add(data, data)

            np.testing.assert_allclose(expected, actual)
            self.assertEqual(actual.dtype, ty)

    def test_nd(self):
        @vectorize(signatures, target="cuda")
        def vector_add(a, b):
            return a + b

        for nd, dtype, order in product(range(1, 8), dtypes, orders):
            shape = (4,) * nd
            data = np.random.random(shape).astype(dtype)
            data2 = np.array(data.T, order=order)

            expected = data + data2
            actual = vector_add(data, data2)
            np.testing.assert_allclose(expected, actual)
            self.assertEqual(actual.dtype, dtype)

    def test_output_arg(self):
        @vectorize(signatures, target="cuda")
        def vector_add(a, b):
            return a + b

        A = np.arange(10, dtype=np.float32)
        B = np.arange(10, dtype=np.float32)

        expected = A + B
        actual = np.empty_like(A)
        vector_add(A, B, out=actual)

        np.testing.assert_allclose(expected, actual)
        self.assertEqual(expected.dtype, actual.dtype)

    def test_reduce(self):
        @vectorize(signatures, target="cuda")
        def vector_add(a, b):
            return a + b

        dtype = np.int32

        for n in input_sizes:
            x = np.arange(n, dtype=dtype)
            expected = np.add.reduce(x)
            actual = vector_add.reduce(x)
            np.testing.assert_allclose(expected, actual)
            # np.add.reduce is special-cased to return an int64 for any int
            # arguments, so we can't compare against its returned dtype when
            # we're checking the general reduce machinery (which just happens
            # to be using addition). Instead, compare against the input dtype.
            self.assertEqual(dtype, actual.dtype)

    def test_reduce_async(self):
        @vectorize(signatures, target="cuda")
        def vector_add(a, b):
            return a + b

        stream = cuda.stream()
        dtype = np.int32

        for n in input_sizes:
            x = np.arange(n, dtype=dtype)
            expected = np.add.reduce(x)
            dx = cuda.to_device(x, stream)
            actual = vector_add.reduce(dx, stream=stream)
            np.testing.assert_allclose(expected, actual)
            # Compare against the input dtype as in test_reduce().
            self.assertEqual(dtype, actual.dtype)

    def test_manual_transfer(self):
        @vectorize(signatures, target="cuda")
        def vector_add(a, b):
            return a + b

        n = 10
        x = np.arange(n, dtype=np.int32)
        dx = cuda.to_device(x)
        expected = x + x
        actual = vector_add(x, dx).copy_to_host()
        np.testing.assert_equal(expected, actual)
        self.assertEqual(expected.dtype, actual.dtype)

    def test_ufunc_output_2d(self):
        @vectorize(signatures, target="cuda")
        def vector_add(a, b):
            return a + b

        n = 10
        x = np.arange(n, dtype=np.int32).reshape(2, 5)
        dx = cuda.to_device(x)
        vector_add(dx, dx, out=dx)

        expected = x + x
        actual = dx.copy_to_host()
        np.testing.assert_equal(expected, actual)
        self.assertEqual(expected.dtype, actual.dtype)

    def check_tuple_arg(self, a, b):
        @vectorize(signatures, target="cuda")
        def vector_add(a, b):
            return a + b

        r = vector_add(a, b)
        np.testing.assert_equal(np.asarray(a) + np.asarray(b), r)

    def test_tuple_arg(self):
        a = (1.0, 2.0, 3.0)
        b = (4.0, 5.0, 6.0)
        self.check_tuple_arg(a, b)

    def test_namedtuple_arg(self):
        Point = namedtuple("Point", ("x", "y", "z"))
        a = Point(x=1.0, y=2.0, z=3.0)
        b = Point(x=4.0, y=5.0, z=6.0)
        self.check_tuple_arg(a, b)

    def test_tuple_of_array_arg(self):
        arr = np.arange(10, dtype=np.int32)
        a = (arr, arr + 1)
        b = (arr + 2, arr + 2)
        self.check_tuple_arg(a, b)

    def test_tuple_of_namedtuple_arg(self):
        Point = namedtuple("Point", ("x", "y", "z"))
        a = (Point(x=1.0, y=2.0, z=3.0), Point(x=1.5, y=2.5, z=3.5))
        b = (Point(x=4.0, y=5.0, z=6.0), Point(x=4.5, y=5.5, z=6.5))
        self.check_tuple_arg(a, b)

    def test_namedtuple_of_array_arg(self):
        xs1 = np.arange(10, dtype=np.int32)
        ys1 = xs1 + 2
        xs2 = np.arange(10, dtype=np.int32) * 2
        ys2 = xs2 + 1
        Points = namedtuple("Points", ("xs", "ys"))
        a = Points(xs=xs1, ys=ys1)
        b = Points(xs=xs2, ys=ys2)
        self.check_tuple_arg(a, b)

    def test_name_attribute(self):
        @vectorize("f8(f8)", target="cuda")
        def bar(x):
            return x**2

        self.assertEqual(bar.__name__, "bar")

    def test_no_transfer_for_device_data(self):
        # Initialize test data on the device prior to banning host <-> device
        # transfer

        noise = np.random.randn(1, 3, 64, 64).astype(np.float32)
        noise = cuda.to_device(noise)

        # A mock of a CUDA function that always raises a CudaAPIError

        def raising_transfer(*args, **kwargs):
            raise CudaAPIError(999, "Transfer not allowed")

        # Use the mock for transfers between the host and device

        old_HtoD = getattr(driver, "cuMemcpyHtoD", None)
        old_DtoH = getattr(driver, "cuMemcpyDtoH", None)

        setattr(driver, "cuMemcpyHtoD", raising_transfer)
        setattr(driver, "cuMemcpyDtoH", raising_transfer)

        # Ensure that the mock functions are working as expected

        with self.assertRaisesRegex(CudaAPIError, "Transfer not allowed"):
            noise.copy_to_host()

        with self.assertRaisesRegex(CudaAPIError, "Transfer not allowed"):
            cuda.to_device([1])

        try:
            # Check that defining and calling a ufunc with data on the device
            # induces no transfers

            @vectorize(["float32(float32)"], target="cuda")
            def func(noise):
                return noise + 1.0

            func(noise)
        finally:
            # Replace our mocks with the original implementations. If there was
            # no original implementation, simply remove ours.

            if old_HtoD is not None:
                setattr(driver, "cuMemcpyHtoD", old_HtoD)
            else:
                del driver.cuMemcpyHtoD
            if old_DtoH is not None:
                setattr(driver, "cuMemcpyDtoH", old_DtoH)
            else:
                del driver.cuMemcpyDtoH


if __name__ == "__main__":
    unittest.main()



from numba.cuda.testing import (
    skip_on_cudasim,
    skip_unless_cudasim,
    unittest,
    CUDATestCase,
)
from numba import config, cuda

# Basic tests that stream APIs execute on the hardware and in the simulator.
#
# Correctness of semantics is exercised elsewhere in the test suite (though we
# could improve the comprehensiveness of testing by adding more correctness
# tests here in future).


class TestStreamAPI(CUDATestCase):
    def test_stream_create_and_sync(self):
        s = cuda.stream()
        s.synchronize()

    def test_default_stream_create_and_sync(self):
        s = cuda.default_stream()
        s.synchronize()

    def test_legacy_default_stream_create_and_sync(self):
        s = cuda.legacy_default_stream()
        s.synchronize()

    def test_ptd_stream_create_and_sync(self):
        s = cuda.per_thread_default_stream()
        s.synchronize()

    @skip_on_cudasim("External streams are unsupported on the simulator")
    def test_external_stream_create(self):
        #  A dummy pointer value
        ptr = 0x12345678
        s = cuda.external_stream(ptr)
        # We don't test synchronization on the stream because it's not a real
        # stream - we used a dummy pointer for testing the API, so we just
        # ensure that the stream handle matches the external stream pointer.
        if config.CUDA_USE_NVIDIA_BINDING:
            value = int(s.handle)
        else:
            value = s.handle.value
        self.assertEqual(ptr, value)

    @skip_unless_cudasim("External streams are usable with hardware")
    def test_external_stream_simulator_unavailable(self):
        ptr = 0x12345678
        msg = "External streams are unsupported in the simulator"
        with self.assertRaisesRegex(RuntimeError, msg):
            cuda.external_stream(ptr)


if __name__ == "__main__":
    unittest.main()



"""CUDA Driver

- Driver API binding
- NVVM API binding
- Device array implementation

"""

from numba.core import config

assert not config.ENABLE_CUDASIM, "Cannot use real driver API with simulator"



import numpy as np

from numba.core.utils import PYVERSION
from numba.cuda.testing import skip_on_cudasim, CUDATestCase
from numba.tests.support import (
    override_config,
    captured_stderr,
    captured_stdout,
)
from numba import cuda, float64
import unittest


def simple_cuda(A, B):
    i = cuda.grid(1)
    B[i] = A[i] + 1.5


@skip_on_cudasim("Simulator does not produce debug dumps")
class TestDebugOutput(CUDATestCase):
    def compile_simple_cuda(self):
        with captured_stderr() as err:
            with captured_stdout() as out:
                cfunc = cuda.jit((float64[:], float64[:]))(simple_cuda)
                # Call compiled function (to ensure PTX is generated)
                # and sanity-check results.
                A = np.linspace(0, 1, 10).astype(np.float64)
                B = np.zeros_like(A)
                cfunc[1, 10](A, B)
                self.assertTrue(np.allclose(A + 1.5, B))
        # stderr shouldn't be affected by debug output
        self.assertFalse(err.getvalue())
        return out.getvalue()

    def assert_fails(self, *args, **kwargs):
        self.assertRaises(AssertionError, *args, **kwargs)

    def check_debug_output(self, out, enabled_dumps):
        all_dumps = dict.fromkeys(
            ["bytecode", "cfg", "ir", "llvm", "assembly"], False
        )
        for name in enabled_dumps:
            assert name in all_dumps
            all_dumps[name] = True
        for name, enabled in sorted(all_dumps.items()):
            check_meth = getattr(self, "_check_dump_%s" % name)
            if enabled:
                check_meth(out)
            else:
                self.assertRaises(AssertionError, check_meth, out)

    def _check_dump_bytecode(self, out):
        if PYVERSION > (3, 10):
            # binop with arg=0 is binary add, see CPython dis.py and opcode.py
            self.assertIn("BINARY_OP(arg=0", out)
        else:
            self.assertIn("BINARY_ADD", out)

    def _check_dump_cfg(self, out):
        self.assertIn("CFG dominators", out)

    def _check_dump_ir(self, out):
        self.assertIn("--IR DUMP: simple_cuda--", out)
        self.assertIn("const(float, 1.5)", out)

    def _check_dump_llvm(self, out):
        self.assertIn("--LLVM DUMP", out)
        self.assertIn('!"kernel", i32 1', out)

    def _check_dump_assembly(self, out):
        self.assertIn("--ASSEMBLY simple_cuda", out)
        self.assertIn("Generated by NVIDIA NVVM Compiler", out)

    def test_dump_bytecode(self):
        with override_config("DUMP_BYTECODE", True):
            out = self.compile_simple_cuda()
        self.check_debug_output(out, ["bytecode"])

    def test_dump_ir(self):
        with override_config("DUMP_IR", True):
            out = self.compile_simple_cuda()
        self.check_debug_output(out, ["ir"])

    def test_dump_cfg(self):
        with override_config("DUMP_CFG", True):
            out = self.compile_simple_cuda()
        self.check_debug_output(out, ["cfg"])

    def test_dump_llvm(self):
        with override_config("DUMP_LLVM", True):
            out = self.compile_simple_cuda()
        self.check_debug_output(out, ["llvm"])

    def test_dump_assembly(self):
        with override_config("DUMP_ASSEMBLY", True):
            out = self.compile_simple_cuda()
        self.check_debug_output(out, ["assembly"])


if __name__ == "__main__":
    unittest.main()



from numba.cuda.testing import skip_on_cudasim, unittest, CUDATestCase
from llvmlite import ir

import numpy as np
import os
from numba import config, cuda, njit, types
from numba.extending import overload


class Interval:
    """
    A half-open interval on the real number line.
    """

    def __init__(self, lo, hi):
        self.lo = lo
        self.hi = hi

    def __repr__(self):
        return "Interval(%f, %f)" % (self.lo, self.hi)

    @property
    def width(self):
        return self.hi - self.lo


@njit
def interval_width(interval):
    return interval.width


@njit
def sum_intervals(i, j):
    return Interval(i.lo + j.lo, i.hi + j.hi)


if not config.ENABLE_CUDASIM:
    from numba.core import cgutils
    from numba.core.extending import (
        lower_builtin,
        models,
        type_callable,
        typeof_impl,
    )
    from numba.core.typing.templates import AttributeTemplate
    from numba.cuda.cudadecl import registry as cuda_registry
    from numba.cuda.cudaimpl import lower_attr as cuda_lower_attr
    from numba.cuda.extending import (
        register_model,
        make_attribute_wrapper,
    )

    class IntervalType(types.Type):
        def __init__(self):
            super().__init__(name="Interval")

    interval_type = IntervalType()

    @typeof_impl.register(Interval)
    def typeof_interval(val, c):
        return interval_type

    @type_callable(Interval)
    def type_interval(context):
        def typer(lo, hi):
            if isinstance(lo, types.Float) and isinstance(hi, types.Float):
                return interval_type

        return typer

    @register_model(IntervalType)
    class IntervalModel(models.StructModel):
        def __init__(self, dmm, fe_type):
            members = [
                ("lo", types.float64),
                ("hi", types.float64),
            ]
            models.StructModel.__init__(self, dmm, fe_type, members)

    make_attribute_wrapper(IntervalType, "lo", "lo")
    make_attribute_wrapper(IntervalType, "hi", "hi")

    @lower_builtin(Interval, types.Float, types.Float)
    def impl_interval(context, builder, sig, args):
        typ = sig.return_type
        lo, hi = args
        interval = cgutils.create_struct_proxy(typ)(context, builder)
        interval.lo = lo
        interval.hi = hi
        return interval._getvalue()

    @cuda_registry.register_attr
    class Interval_attrs(AttributeTemplate):
        key = IntervalType

        def resolve_width(self, mod):
            return types.float64

    @cuda_lower_attr(IntervalType, "width")
    def cuda_Interval_width(context, builder, sig, arg):
        lo = builder.extract_value(arg, 0)
        hi = builder.extract_value(arg, 1)
        return builder.fsub(hi, lo)


@skip_on_cudasim("Extensions not supported in the simulator")
class TestExtending(CUDATestCase):
    def test_attributes(self):
        @cuda.jit
        def f(r, x):
            iv = Interval(x[0], x[1])
            r[0] = iv.lo
            r[1] = iv.hi

        x = np.asarray((1.5, 2.5))
        r = np.zeros_like(x)

        f[1, 1](r, x)

        np.testing.assert_equal(r, x)

    def test_property(self):
        @cuda.jit
        def f(r, x):
            iv = Interval(x[0], x[1])
            r[0] = iv.width

        x = np.asarray((1.5, 2.5))
        r = np.zeros(1)

        f[1, 1](r, x)

        np.testing.assert_allclose(r[0], x[1] - x[0])

    def test_extension_type_as_arg(self):
        @cuda.jit
        def f(r, x):
            iv = Interval(x[0], x[1])
            r[0] = interval_width(iv)

        x = np.asarray((1.5, 2.5))
        r = np.zeros(1)

        f[1, 1](r, x)

        np.testing.assert_allclose(r[0], x[1] - x[0])

    def test_extension_type_as_retvalue(self):
        @cuda.jit
        def f(r, x):
            iv1 = Interval(x[0], x[1])
            iv2 = Interval(x[2], x[3])
            iv_sum = sum_intervals(iv1, iv2)
            r[0] = iv_sum.lo
            r[1] = iv_sum.hi

        x = np.asarray((1.5, 2.5, 3.0, 4.0))
        r = np.zeros(2)

        f[1, 1](r, x)

        expected = np.asarray((x[0] + x[2], x[1] + x[3]))
        np.testing.assert_allclose(r, expected)


TEST_BIN_DIR = os.getenv("NUMBA_CUDA_TEST_BIN_DIR")
if TEST_BIN_DIR:
    test_device_functions_a = os.path.join(
        TEST_BIN_DIR, "test_device_functions.a"
    )
    test_device_functions_cubin = os.path.join(
        TEST_BIN_DIR, "test_device_functions.cubin"
    )
    test_device_functions_cu = os.path.join(
        TEST_BIN_DIR, "test_device_functions.cu"
    )
    test_device_functions_fatbin = os.path.join(
        TEST_BIN_DIR, "test_device_functions.fatbin"
    )
    test_device_functions_fatbin_multi = os.path.join(
        TEST_BIN_DIR, "test_device_functions_multi.fatbin"
    )
    test_device_functions_o = os.path.join(
        TEST_BIN_DIR, "test_device_functions.o"
    )
    test_device_functions_ptx = os.path.join(
        TEST_BIN_DIR, "test_device_functions.ptx"
    )
    test_device_functions_ltoir = os.path.join(
        TEST_BIN_DIR, "test_device_functions.ltoir"
    )


class TestExtendingLinkage(CUDATestCase):
    def test_extension_adds_linkable_code(self):
        cuda_major_version = cuda.runtime.get_version()[0]

        if cuda_major_version < 12:
            self.skipTest("CUDA 12 required for linking in-memory data")

        files = (
            (test_device_functions_a, cuda.Archive),
            (test_device_functions_cubin, cuda.Cubin),
            (test_device_functions_cu, cuda.CUSource),
            (test_device_functions_fatbin, cuda.Fatbin),
            (test_device_functions_o, cuda.Object),
            (test_device_functions_ptx, cuda.PTXSource),
            (test_device_functions_ltoir, cuda.LTOIR),
        )

        lto = config.CUDA_ENABLE_PYNVJITLINK

        for path, ctor in files:
            if ctor == cuda.LTOIR and not lto:
                # Don't try to test with LTOIR if LTO is not enabled
                continue

            with open(path, "rb") as f:
                code_object = ctor(f.read())

            def external_add(x, y):
                return x + y

            @type_callable(external_add)
            def type_external_add(context):
                def typer(x, y):
                    if x == types.uint32 and y == types.uint32:
                        return types.uint32

                return typer

            @lower_builtin(external_add, types.uint32, types.uint32)
            def lower_external_add(context, builder, sig, args):
                context.active_code_library.add_linking_file(code_object)
                i32 = ir.IntType(32)
                fnty = ir.FunctionType(i32, [i32, i32])
                fn = cgutils.get_or_insert_function(
                    builder.module, fnty, "add_cabi"
                )
                return builder.call(fn, args)

            @cuda.jit(lto=lto)
            def use_external_add(r, x, y):
                r[0] = external_add(x[0], y[0])

            r = np.zeros(1, dtype=np.uint32)
            x = np.ones(1, dtype=np.uint32)
            y = np.ones(1, dtype=np.uint32) * 2

            use_external_add[1, 1](r, x, y)

            np.testing.assert_equal(r[0], 3)

            @cuda.jit(lto=lto)
            def use_external_add_device(x, y):
                return external_add(x, y)

            @cuda.jit(lto=lto)
            def use_external_add_kernel(r, x, y):
                r[0] = use_external_add_device(x[0], y[0])

            r = np.zeros(1, dtype=np.uint32)
            x = np.ones(1, dtype=np.uint32)
            y = np.ones(1, dtype=np.uint32) * 2

            use_external_add_kernel[1, 1](r, x, y)

            np.testing.assert_equal(r[0], 3)

    def test_linked_called_through_overload(self):
        cu_code = cuda.CUSource("""
            extern "C" __device__
            int bar(int *out, int a)
            {
              *out = a * 2;
              return 0;
            }
        """)

        bar = cuda.declare_device("bar", "int32(int32)", link=cu_code)

        def bar_call(val):
            pass

        @overload(bar_call, target="cuda")
        def ol_bar_call(a):
            return lambda a: bar(a)

        @cuda.jit("void(int32[::1], int32[::1])")
        def foo(r, x):
            i = cuda.grid(1)
            if i < len(r):
                r[i] = bar_call(x[i])

        x = np.arange(10, dtype=np.int32)
        r = np.empty_like(x)

        foo[1, 32](r, x)

        np.testing.assert_equal(r, x * 2)


if __name__ == "__main__":
    unittest.main()



import numpy as np
from numba import cuda
from numba.cuda.testing import unittest, CUDATestCase, skip_on_cudasim
from numba.tests.support import linux_only, override_config
from numba.core.errors import NumbaPerformanceWarning
from numba.core import config
import warnings


@skip_on_cudasim("cudasim does not raise performance warnings")
class TestWarnings(CUDATestCase):
    def test_inefficient_launch_configuration(self):
        @cuda.jit
        def kernel():
            pass

        with override_config("CUDA_LOW_OCCUPANCY_WARNINGS", 1):
            with warnings.catch_warnings(record=True) as w:
                kernel[1, 1]()

        self.assertEqual(w[0].category, NumbaPerformanceWarning)
        self.assertIn("Grid size", str(w[0].message))
        self.assertIn("low occupancy", str(w[0].message))

    def test_efficient_launch_configuration(self):
        @cuda.jit
        def kernel():
            pass

        with override_config("CUDA_LOW_OCCUPANCY_WARNINGS", 1):
            with warnings.catch_warnings(record=True) as w:
                kernel[256, 256]()

        self.assertEqual(len(w), 0)

    def test_warn_on_host_array(self):
        @cuda.jit
        def foo(r, x):
            r[0] = x + 1

        N = 10
        arr_f32 = np.zeros(N, dtype=np.float32)
        with override_config("CUDA_WARN_ON_IMPLICIT_COPY", 1):
            with warnings.catch_warnings(record=True) as w:
                foo[1, N](arr_f32, N)

        self.assertEqual(w[0].category, NumbaPerformanceWarning)
        self.assertIn(
            "Host array used in CUDA kernel will incur", str(w[0].message)
        )
        self.assertIn("copy overhead", str(w[0].message))

    def test_pinned_warn_on_host_array(self):
        @cuda.jit
        def foo(r, x):
            r[0] = x + 1

        N = 10
        ary = cuda.pinned_array(N, dtype=np.float32)

        with override_config("CUDA_WARN_ON_IMPLICIT_COPY", 1):
            with warnings.catch_warnings(record=True) as w:
                foo[1, N](ary, N)

        self.assertEqual(w[0].category, NumbaPerformanceWarning)
        self.assertIn(
            "Host array used in CUDA kernel will incur", str(w[0].message)
        )
        self.assertIn("copy overhead", str(w[0].message))

    def test_nowarn_on_mapped_array(self):
        @cuda.jit
        def foo(r, x):
            r[0] = x + 1

        N = 10
        ary = cuda.mapped_array(N, dtype=np.float32)

        with override_config("CUDA_WARN_ON_IMPLICIT_COPY", 1):
            with warnings.catch_warnings(record=True) as w:
                foo[1, N](ary, N)

        self.assertEqual(len(w), 0)

    @linux_only
    def test_nowarn_on_managed_array(self):
        @cuda.jit
        def foo(r, x):
            r[0] = x + 1

        N = 10
        ary = cuda.managed_array(N, dtype=np.float32)

        with override_config("CUDA_WARN_ON_IMPLICIT_COPY", 1):
            with warnings.catch_warnings(record=True) as w:
                foo[1, N](ary, N)

        self.assertEqual(len(w), 0)

    def test_nowarn_on_device_array(self):
        @cuda.jit
        def foo(r, x):
            r[0] = x + 1

        N = 10
        ary = cuda.device_array(N, dtype=np.float32)

        with override_config("CUDA_WARN_ON_IMPLICIT_COPY", 1):
            with warnings.catch_warnings(record=True) as w:
                foo[1, N](ary, N)

        self.assertEqual(len(w), 0)

    def test_warn_on_debug_and_opt(self):
        with warnings.catch_warnings(record=True) as w:
            cuda.jit(debug=True, opt=True)

        self.assertEqual(len(w), 1)
        self.assertIn("not supported by CUDA", str(w[0].message))

    def test_warn_on_debug_and_opt_default(self):
        with warnings.catch_warnings(record=True) as w:
            cuda.jit(debug=True)

        self.assertEqual(len(w), 1)
        self.assertIn("not supported by CUDA", str(w[0].message))

    def test_no_warn_on_debug_and_no_opt(self):
        with warnings.catch_warnings(record=True) as w:
            cuda.jit(debug=True, opt=False)

        self.assertEqual(len(w), 0)

    def test_no_warn_with_no_debug_and_opt_kwargs(self):
        with warnings.catch_warnings(record=True) as w:
            cuda.jit()

        self.assertEqual(len(w), 0)

    def test_no_warn_on_debug_and_opt_with_config(self):
        with override_config("CUDA_DEBUGINFO_DEFAULT", 1):
            with override_config("OPT", config._OptLevel(0)):
                with warnings.catch_warnings(record=True) as w:
                    cuda.jit()

            self.assertEqual(len(w), 0)

            with warnings.catch_warnings(record=True) as w:
                cuda.jit(opt=False)

            self.assertEqual(len(w), 0)

        with override_config("OPT", config._OptLevel(0)):
            with warnings.catch_warnings(record=True) as w:
                cuda.jit(debug=True)

            self.assertEqual(len(w), 0)

    def test_warn_on_debug_and_opt_with_config(self):
        with override_config("CUDA_DEBUGINFO_DEFAULT", 1):
            for opt in (1, 2, 3, "max"):
                with override_config("OPT", config._OptLevel(opt)):
                    with warnings.catch_warnings(record=True) as w:
                        cuda.jit()

                self.assertEqual(len(w), 1)
                self.assertIn("not supported by CUDA", str(w[0].message))

        for opt in (1, 2, 3, "max"):
            with override_config("OPT", config._OptLevel(opt)):
                with warnings.catch_warnings(record=True) as w:
                    cuda.jit(debug=True)

                self.assertEqual(len(w), 1)
                self.assertIn("not supported by CUDA", str(w[0].message))


if __name__ == "__main__":
    unittest.main()



import multiprocessing as mp
import os

from numba import cuda
from numba.cuda.cudadrv.driver import CudaAPIError, driver
from numba.cuda.cudadrv.error import CudaSupportError
from numba.cuda.testing import skip_on_cudasim, unittest, CUDATestCase


# A mock of cuInit that always raises a CudaAPIError
def cuInit_raising(arg):
    raise CudaAPIError(999, "CUDA_ERROR_UNKNOWN")


# Test code to run in a child that patches driver.cuInit to a variant that
# always raises. We can't use mock.patch.object here because driver.cuInit is
# not assigned until we attempt to initialize - mock.patch.object cannot locate
# the non-existent original method, and so fails. Instead we patch
# driver.cuInit with our raising version prior to any attempt to initialize.
def cuInit_raising_test(result_queue):
    driver.cuInit = cuInit_raising

    success = False
    msg = None

    try:
        # A CUDA operation that forces initialization of the device
        cuda.device_array(1)
    except CudaSupportError as e:
        success = True
        msg = e.msg

    result_queue.put((success, msg))


# Similar to cuInit_raising_test above, but for testing that the string
# returned by cuda_error() is as expected.
def initialization_error_test(result_queue):
    driver.cuInit = cuInit_raising

    success = False
    msg = None

    try:
        # A CUDA operation that forces initialization of the device
        cuda.device_array(1)
    except CudaSupportError:
        success = True

    msg = cuda.cuda_error()
    result_queue.put((success, msg))


# For testing the path where Driver.__init__() catches a CudaSupportError
def cuda_disabled_test(result_queue):
    success = False
    msg = None

    try:
        # A CUDA operation that forces initialization of the device
        cuda.device_array(1)
    except CudaSupportError as e:
        success = True
        msg = e.msg

    result_queue.put((success, msg))


# Similar to cuda_disabled_test, but checks cuda.cuda_error() instead of the
# exception raised on initialization
def cuda_disabled_error_test(result_queue):
    success = False
    msg = None

    try:
        # A CUDA operation that forces initialization of the device
        cuda.device_array(1)
    except CudaSupportError:
        success = True

    msg = cuda.cuda_error()
    result_queue.put((success, msg))


@skip_on_cudasim("CUDA Simulator does not initialize driver")
class TestInit(CUDATestCase):
    def _test_init_failure(self, target, expected):
        # Run the initialization failure test in a separate subprocess
        ctx = mp.get_context("spawn")
        result_queue = ctx.Queue()
        proc = ctx.Process(target=target, args=(result_queue,))
        proc.start()
        proc.join(30)  # should complete within 30s
        success, msg = result_queue.get()

        # Ensure the child process raised an exception during initialization
        # before checking the message
        if not success:
            self.fail("CudaSupportError not raised")

        self.assertIn(expected, msg)

    def test_init_failure_raising(self):
        expected = "Error at driver init: CUDA_ERROR_UNKNOWN (999)"
        self._test_init_failure(cuInit_raising_test, expected)

    def test_init_failure_error(self):
        expected = "CUDA_ERROR_UNKNOWN (999)"
        self._test_init_failure(initialization_error_test, expected)

    def _test_cuda_disabled(self, target):
        # Uses _test_init_failure to launch the test in a separate subprocess
        # with CUDA disabled.
        cuda_disabled = os.environ.get("NUMBA_DISABLE_CUDA")
        os.environ["NUMBA_DISABLE_CUDA"] = "1"
        try:
            expected = "CUDA is disabled due to setting NUMBA_DISABLE_CUDA=1"
            self._test_init_failure(cuda_disabled_test, expected)
        finally:
            if cuda_disabled is not None:
                os.environ["NUMBA_DISABLE_CUDA"] = cuda_disabled
            else:
                os.environ.pop("NUMBA_DISABLE_CUDA")

    def test_cuda_disabled_raising(self):
        self._test_cuda_disabled(cuda_disabled_test)

    def test_cuda_disabled_error(self):
        self._test_cuda_disabled(cuda_disabled_error_test)

    def test_init_success(self):
        # Here we assume that initialization is successful (because many bad
        # things will happen with the test suite if it is not) and check that
        # there is no error recorded.
        self.assertIsNone(cuda.cuda_error())


if __name__ == "__main__":
    unittest.main()



import os
import sys
import subprocess
import threading
from numba import cuda
from numba.cuda.testing import (
    unittest,
    CUDATestCase,
    skip_on_cudasim,
    skip_under_cuda_memcheck,
)
from numba.tests.support import captured_stdout


class TestCudaDetect(CUDATestCase):
    def test_cuda_detect(self):
        # exercise the code path
        with captured_stdout() as out:
            cuda.detect()
        output = out.getvalue()
        self.assertIn("Found", output)
        self.assertIn("CUDA devices", output)


@skip_under_cuda_memcheck("Hangs cuda-memcheck")
class TestCUDAFindLibs(CUDATestCase):
    def run_cmd(self, cmdline, env):
        popen = subprocess.Popen(
            cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env
        )

        # finish in 5 minutes or kill it
        timeout = threading.Timer(5 * 60.0, popen.kill)
        try:
            timeout.start()
            out, err = popen.communicate()
            # the process should exit with an error
            return out.decode(), err.decode()
        finally:
            timeout.cancel()
        return None, None

    def run_test_in_separate_process(self, envvar, envvar_value):
        env_copy = os.environ.copy()
        env_copy[envvar] = str(envvar_value)
        code = """if 1:
            from numba import cuda
            @cuda.jit('(int64,)')
            def kernel(x):
                pass
            kernel(1,)
            """
        cmdline = [sys.executable, "-c", code]
        return self.run_cmd(cmdline, env_copy)

    @skip_on_cudasim("Simulator does not hit device library search code path")
    @unittest.skipIf(not sys.platform.startswith("linux"), "linux only")
    def test_cuda_find_lib_errors(self):
        """
        This tests that the find_libs works as expected in the case of an
        environment variable being used to set the path.
        """
        # one of these is likely to exist on linux, it's also unlikely that
        # someone has extracted the contents of libdevice into here!
        locs = ["lib", "lib64"]

        looking_for = None
        for l in locs:
            looking_for = os.path.join(os.path.sep, l)
            if os.path.exists(looking_for):
                break

        # This is the testing part, the test will only run if there's a valid
        # path in which to look
        if looking_for is not None:
            out, err = self.run_test_in_separate_process(
                "NUMBA_CUDA_DRIVER", looking_for
            )
            self.assertTrue(out is not None)
            self.assertTrue(err is not None)


if __name__ == "__main__":
    unittest.main()



import numba.cuda as cuda
from numba.cuda.testing import unittest, CUDATestCase
import numpy as np

from numba import int16, int32, int64, uint16, uint32, uint64, float32, float64
from numba.types import float16

from numba.cuda.cuda_bf16 import (
    nv_bfloat16,
    htrunc,
    hceil,
    hfloor,
    hrint,
    hsqrt,
    hrsqrt,
    hrcp,
    hlog,
    hlog2,
    hlog10,
    hcos,
    hsin,
    hexp,
    hexp2,
    hexp10,
)

from numba.cuda.cudadrv.runtime import get_version

cuda_version = get_version()

dtypes = [int16, int32, int64, uint16, uint32, uint64, float32]


@unittest.skipIf(
    (cuda.get_current_device().compute_capability < (8, 0)),
    "bfloat16 requires compute capability 8.0+",
)
class Bfloat16Test(CUDATestCase):
    def test_ctor(self):
        @cuda.jit
        def simple_kernel():
            a = nv_bfloat16(float64(1.0))  # noqa: F841
            b = nv_bfloat16(float32(2.0))  # noqa: F841
            c = nv_bfloat16(int16(3))  # noqa: F841
            d = nv_bfloat16(int32(4))  # noqa: F841
            e = nv_bfloat16(int64(5))  # noqa: F841
            f = nv_bfloat16(uint16(6))  # noqa: F841
            g = nv_bfloat16(uint32(7))  # noqa: F841
            h = nv_bfloat16(uint64(8))  # noqa: F841

        simple_kernel[1, 1]()

        if cuda_version >= (12, 0):

            @cuda.jit
            def simple_kernel_fp16():
                i = nv_bfloat16(float16(9))  # noqa: F841

            simple_kernel_fp16[1, 1]()

    def test_casts(self):
        @cuda.jit
        def simple_kernel(b, c, d, e, f, g, h):
            a = nv_bfloat16(3.14)

            b[0] = float32(a)
            c[0] = int16(a)
            d[0] = int32(a)
            e[0] = int64(a)
            f[0] = uint16(a)
            g[0] = uint32(a)
            h[0] = uint64(a)

        b = np.zeros(1, dtype=np.float32)
        c = np.zeros(1, dtype=np.int16)
        d = np.zeros(1, dtype=np.int32)
        e = np.zeros(1, dtype=np.int64)
        f = np.zeros(1, dtype=np.uint16)
        g = np.zeros(1, dtype=np.uint32)
        h = np.zeros(1, dtype=np.uint64)

        simple_kernel[1, 1](b, c, d, e, f, g, h)

        np.testing.assert_allclose(b[0], 3.14, atol=1e-2)
        assert c[0] == 3
        assert d[0] == 3
        assert e[0] == 3
        assert f[0] == 3
        assert g[0] == 3
        assert h[0] == 3

    def test_ctor_cast_loop(self):
        for dtype in dtypes:
            with self.subTest(dtype=dtype):

                @cuda.jit
                def simple_kernel(a):
                    a[0] = dtype(nv_bfloat16(dtype(3.14)))

                a = np.zeros(1, dtype=str(dtype))
                simple_kernel[1, 1](a)

                if np.dtype(str(dtype)).kind == "f":
                    np.testing.assert_allclose(a[0], 3.14, atol=1e-2)
                else:
                    assert a[0] == 3

    def test_arithmetic(self):
        @cuda.jit
        def simple_kernel(arith, logic):
            # Binary Arithmetic Operators
            a = nv_bfloat16(1.0)
            b = nv_bfloat16(2.0)

            arith[0] = float32(a + b)
            arith[1] = float32(a - b)
            arith[2] = float32(a * b)
            arith[3] = float32(a / b)

            # Arithmetic Assignment Operators
            a = nv_bfloat16(1.0)
            b = nv_bfloat16(2.0)

            a += b
            arith[4] = float32(a)
            a -= b
            arith[5] = float32(a)
            a *= b
            arith[6] = float32(a)
            a /= b
            arith[7] = float32(a)

            # Unary Arithmetic Operators
            a = nv_bfloat16(1.0)

            arith[8] = float32(+a)
            arith[9] = float32(-a)

            # Comparison Operators
            a = nv_bfloat16(1.0)
            b = nv_bfloat16(2.0)

            logic[0] = a == b
            logic[1] = a != b
            logic[2] = a > b
            logic[3] = a < b
            logic[4] = a >= b
            logic[5] = a <= b

        arith = np.zeros(10, dtype=np.float32)
        logic = np.zeros(6, dtype=np.bool_)

        simple_kernel[1, 1](arith, logic)

        a = 1.0
        b = 2.0
        np.testing.assert_allclose(
            arith,
            [
                a + b,
                a - b,
                a * b,
                a / b,
                a + b,
                a + b - b,
                (a + b - b) * b,
                (a + b - b) * b / b,
                +a,
                -a,
            ],
            atol=1e-2,
        )
        np.testing.assert_equal(
            logic, [a == b, a != b, a > b, a < b, a >= b, a <= b]
        )

    def test_math_func(self):
        @cuda.jit
        def simple_kernel(a):
            x = nv_bfloat16(3.14)

            a[0] = float32(htrunc(x))
            a[1] = float32(hceil(x))
            a[2] = float32(hfloor(x))
            a[3] = float32(hrint(x))
            a[4] = float32(hsqrt(x))
            a[5] = float32(hrsqrt(x))
            a[6] = float32(hrcp(x))
            a[7] = float32(hlog(x))
            a[8] = float32(hlog2(x))
            a[9] = float32(hlog10(x))
            a[10] = float32(hcos(x))
            a[11] = float32(hsin(x))
            a[12] = float32(hexp(x))
            a[13] = float32(hexp2(x))
            a[14] = float32(hexp10(x))

        a = np.zeros(15, dtype=np.float32)
        simple_kernel[1, 1](a)

        x = 3.14
        np.testing.assert_allclose(
            a[:12],
            [
                np.trunc(x),
                np.ceil(x),
                np.floor(x),
                np.rint(x),
                np.sqrt(x),
                1 / np.sqrt(x),
                1 / x,
                np.log(x),
                np.log2(x),
                np.log10(x),
                np.cos(x),
                np.sin(x),
            ],
            atol=1e-2,
        )

        np.testing.assert_allclose(
            a[12:], [np.exp(x), np.exp2(x), np.power(10, x)], atol=1e2
        )

    def test_check_bfloat16_type(self):
        @cuda.jit
        def kernel(arr):
            x = nv_bfloat16(3.14)
            if isinstance(x, nv_bfloat16):
                arr[0] = float32(x)
            else:
                arr[0] = float32(0.0)

        arr = np.zeros(1, np.float32)
        kernel[1, 1](arr)

        np.testing.assert_allclose(arr, [3.14], atol=1e-2)

    def test_use_within_device_func(self):
        @cuda.jit(device=True)
        def add_bf16(a, b):
            return a + b

        @cuda.jit
        def kernel(arr):
            a = nv_bfloat16(3.14)
            b = nv_bfloat16(5)
            arr[0] = float32(hfloor(add_bf16(a, b)))

        arr = np.zeros(1, np.float32)
        kernel[1, 1](arr)

        np.testing.assert_allclose(arr, [8], atol=1e-2)


if __name__ == "__main__":
    unittest.main()



from numba import float64, uint32
from numba.cuda.compiler import compile_ptx
from numba.cuda.testing import skip_on_cudasim, unittest


@skip_on_cudasim("Compilation unsupported in the simulator")
class TestCudaMandel(unittest.TestCase):
    def test_mandel(self):
        """Just make sure we can compile this"""

        def mandel(tid, min_x, max_x, min_y, max_y, width, height, iters):
            pixel_size_x = (max_x - min_x) / width
            pixel_size_y = (max_y - min_y) / height

            x = tid % width
            y = tid / width

            real = min_x + x * pixel_size_x
            imag = min_y + y * pixel_size_y

            c = complex(real, imag)
            z = 0.0j

            for i in range(iters):
                z = z * z + c
                if (z.real * z.real + z.imag * z.imag) >= 4:
                    return i
            return iters

        args = (
            uint32,
            float64,
            float64,
            float64,
            float64,
            uint32,
            uint32,
            uint32,
        )
        compile_ptx(mandel, args, device=True)


if __name__ == "__main__":
    unittest.main()



import math
import itertools

import numpy as np

from numba.cuda.testing import unittest, CUDATestCase
from numba.core import types
from numba import cuda
from numba.tests.complex_usecases import (
    real_usecase,
    imag_usecase,
    conjugate_usecase,
    phase_usecase,
    polar_as_complex_usecase,
    rect_usecase,
    isnan_usecase,
    isinf_usecase,
    isfinite_usecase,
    exp_usecase,
    log_usecase,
    log_base_usecase,
    log10_usecase,
    sqrt_usecase,
    asin_usecase,
    acos_usecase,
    atan_usecase,
    cos_usecase,
    sin_usecase,
    tan_usecase,
    acosh_usecase,
    asinh_usecase,
    atanh_usecase,
    cosh_usecase,
    sinh_usecase,
    tanh_usecase,
)
from numba.np import numpy_support


def compile_scalar_func(pyfunc, argtypes, restype):
    # First compile a scalar device function
    assert not any(isinstance(tp, types.Array) for tp in argtypes)
    assert not isinstance(restype, types.Array)
    device_func = cuda.jit(restype(*argtypes), device=True)(pyfunc)

    kernel_types = [
        types.Array(tp, 1, "C") for tp in [restype] + list(argtypes)
    ]

    if len(argtypes) == 1:

        def kernel_func(out, a):
            i = cuda.grid(1)
            if i < out.shape[0]:
                out[i] = device_func(a[i])
    elif len(argtypes) == 2:

        def kernel_func(out, a, b):
            i = cuda.grid(1)
            if i < out.shape[0]:
                out[i] = device_func(a[i], b[i])
    else:
        assert 0

    kernel = cuda.jit(tuple(kernel_types))(kernel_func)

    def kernel_wrapper(values):
        n = len(values)
        inputs = [
            np.empty(n, dtype=numpy_support.as_dtype(tp)) for tp in argtypes
        ]
        output = np.empty(n, dtype=numpy_support.as_dtype(restype))
        for i, vs in enumerate(values):
            for v, inp in zip(vs, inputs):
                inp[i] = v
        args = [output] + inputs
        kernel[int(math.ceil(n / 256)), 256](*args)
        return list(output)

    return kernel_wrapper


class BaseComplexTest(CUDATestCase):
    def basic_values(self):
        reals = [
            -0.0,
            +0.0,
            1,
            -1,
            +1.5,
            -3.5,
            float("-inf"),
            float("+inf"),
            float("nan"),
        ]
        return [complex(x, y) for x, y in itertools.product(reals, reals)]

    def more_values(self):
        reals = [
            0.0,
            +0.0,
            1,
            -1,
            -math.pi,
            +math.pi,
            float("-inf"),
            float("+inf"),
            float("nan"),
        ]
        return [complex(x, y) for x, y in itertools.product(reals, reals)]

    def non_nan_values(self):
        reals = [
            -0.0,
            +0.0,
            1,
            -1,
            -math.pi,
            +math.pi,
            float("inf"),
            float("-inf"),
        ]
        return [complex(x, y) for x, y in itertools.product(reals, reals)]

    def run_func(self, pyfunc, sigs, values, ulps=1, ignore_sign_on_zero=False):
        for sig in sigs:
            if isinstance(sig, types.Type):
                sig = (sig,)
            if isinstance(sig, tuple):
                # Assume return type is the type of first argument
                sig = sig[0](*sig)
            prec = (
                "single"
                if sig.args[0] in (types.float32, types.complex64)
                else "double"
            )
            cudafunc = compile_scalar_func(pyfunc, sig.args, sig.return_type)
            ok_values = []
            expected_list = []
            for args in values:
                if not isinstance(args, (list, tuple)):
                    args = (args,)
                try:
                    expected_list.append(pyfunc(*args))
                    ok_values.append(args)
                except ValueError as e:
                    self.assertIn("math domain error", str(e))
                    continue
            got_list = cudafunc(ok_values)
            for got, expected, args in zip(got_list, expected_list, ok_values):
                msg = "for input %r with prec %r" % (args, prec)
                self.assertPreciseEqual(
                    got,
                    expected,
                    prec=prec,
                    ulps=ulps,
                    ignore_sign_on_zero=ignore_sign_on_zero,
                    msg=msg,
                )

    run_unary = run_func
    run_binary = run_func


class TestComplex(BaseComplexTest):
    def check_real_image(self, pyfunc):
        values = self.basic_values()
        self.run_unary(
            pyfunc,
            [
                tp.underlying_float(tp)
                for tp in (types.complex64, types.complex128)
            ],
            values,
        )

    def test_real(self):
        self.check_real_image(real_usecase)

    def test_imag(self):
        self.check_real_image(imag_usecase)

    def test_conjugate(self):
        pyfunc = conjugate_usecase
        values = self.basic_values()
        self.run_unary(pyfunc, [types.complex64, types.complex128], values)


class TestCMath(BaseComplexTest):
    """
    Tests for cmath module support.
    """

    def check_predicate_func(self, pyfunc):
        self.run_unary(
            pyfunc,
            [types.boolean(tp) for tp in (types.complex128, types.complex64)],
            self.basic_values(),
        )

    def check_unary_func(
        self,
        pyfunc,
        ulps=1,
        values=None,
        returns_float=False,
        ignore_sign_on_zero=False,
    ):
        if returns_float:

            def sig(tp):
                return tp.underlying_float(tp)
        else:

            def sig(tp):
                return tp(tp)

        self.run_unary(
            pyfunc,
            [sig(types.complex128)],
            values or self.more_values(),
            ulps=ulps,
            ignore_sign_on_zero=ignore_sign_on_zero,
        )
        # Avoid discontinuities around pi when in single precision.
        self.run_unary(
            pyfunc,
            [sig(types.complex64)],
            values or self.basic_values(),
            ulps=ulps,
            ignore_sign_on_zero=ignore_sign_on_zero,
        )

    # Conversions

    def test_phase(self):
        self.check_unary_func(phase_usecase, returns_float=True)

    def test_polar(self):
        self.check_unary_func(polar_as_complex_usecase)

    def test_rect(self):
        def do_test(tp, seed_values):
            values = [
                (z.real, z.imag)
                for z in seed_values
                if not math.isinf(z.imag) or z.real == 0
            ]
            float_type = tp.underlying_float
            self.run_binary(rect_usecase, [tp(float_type, float_type)], values)

        do_test(types.complex128, self.more_values())
        # Avoid discontinuities around pi when in single precision.
        do_test(types.complex64, self.basic_values())

    # Classification

    def test_isnan(self):
        self.check_predicate_func(isnan_usecase)

    def test_isinf(self):
        self.check_predicate_func(isinf_usecase)

    def test_isfinite(self):
        self.check_predicate_func(isfinite_usecase)

    # Power and logarithms

    def test_exp(self):
        self.check_unary_func(exp_usecase, ulps=2)

    def test_log(self):
        self.check_unary_func(log_usecase)

    def test_log_base(self):
        values = list(itertools.product(self.more_values(), self.more_values()))
        value_types = [
            (types.complex128, types.complex128),
            (types.complex64, types.complex64),
        ]
        self.run_binary(log_base_usecase, value_types, values, ulps=3)

    def test_log10(self):
        self.check_unary_func(log10_usecase)

    def test_sqrt(self):
        self.check_unary_func(sqrt_usecase)

    # Trigonometric functions

    def test_acos(self):
        self.check_unary_func(acos_usecase, ulps=2)

    def test_asin(self):
        self.check_unary_func(asin_usecase, ulps=2)

    def test_atan(self):
        self.check_unary_func(
            atan_usecase, ulps=2, values=self.non_nan_values()
        )

    def test_cos(self):
        self.check_unary_func(cos_usecase, ulps=2)

    def test_sin(self):
        # See test_sinh.
        self.check_unary_func(sin_usecase, ulps=2)

    def test_tan(self):
        self.check_unary_func(tan_usecase, ulps=2, ignore_sign_on_zero=True)

    # Hyperbolic functions

    def test_acosh(self):
        self.check_unary_func(acosh_usecase)

    def test_asinh(self):
        self.check_unary_func(asinh_usecase, ulps=2)

    def test_atanh(self):
        self.check_unary_func(atanh_usecase, ulps=2, ignore_sign_on_zero=True)

    def test_cosh(self):
        self.check_unary_func(cosh_usecase, ulps=2)

    def test_sinh(self):
        self.check_unary_func(sinh_usecase, ulps=2)

    def test_tanh(self):
        self.check_unary_func(tanh_usecase, ulps=2, ignore_sign_on_zero=True)


class TestAtomicOnComplexComponents(CUDATestCase):
    # Based on the reproducer from Issue #8309. array.real and array.imag could
    # not be used because they required returning an array from a generated
    # function, and even if this was permitted, they could not be resolved from
    # the atomic lowering when they were overloads.
    #
    # See https://github.com/numba/numba/issues/8309

    def test_atomic_on_real(self):
        @cuda.jit
        def atomic_add_one(values):
            i = cuda.grid(1)
            cuda.atomic.add(values.real, i, 1)

        N = 32
        arr1 = np.arange(N) + np.arange(N) * 1j
        arr2 = arr1.copy()
        atomic_add_one[1, N](arr2)
        np.testing.assert_equal(arr1 + 1, arr2)

    def test_atomic_on_imag(self):
        @cuda.jit
        def atomic_add_one_j(values):
            i = cuda.grid(1)
            cuda.atomic.add(values.imag, i, 1)

        N = 32
        arr1 = np.arange(N) + np.arange(N) * 1j
        arr2 = arr1.copy()
        atomic_add_one_j[1, N](arr2)
        np.testing.assert_equal(arr1 + 1j, arr2)


if __name__ == "__main__":
    unittest.main()



from numba import runtests
from numba.core import config

if config.ENABLE_CUDASIM:
    from .simulator_init import *
else:
    from .device_init import *
    from .device_init import _auto_device

from numba.cuda.compiler import (
    compile,
    compile_for_current_device,
    compile_ptx,
    compile_ptx_for_current_device,
)

# This is the out-of-tree NVIDIA-maintained target. This is reported in Numba
# sysinfo (`numba -s`):
implementation = "NVIDIA"


def test(*args, **kwargs):
    if not is_available():
        raise cuda_error()

    return runtests.main("numba.cuda.tests", *args, **kwargs)



import multiprocessing as mp
import traceback
from numba.cuda.testing import unittest, CUDATestCase
from numba.cuda.testing import (
    skip_on_cudasim,
    skip_under_cuda_memcheck,
    skip_if_mvc_libraries_unavailable,
)
from numba.tests.support import linux_only


def child_test():
    from numba import config, cuda

    # Change the MVC config after importing numba.cuda
    config.CUDA_ENABLE_MINOR_VERSION_COMPATIBILITY = 1

    @cuda.jit
    def f():
        pass

    f[1, 1]()


def child_test_wrapper(result_queue):
    try:
        output = child_test()
        success = True
    # Catch anything raised so it can be propagated
    except:  # noqa: E722
        output = traceback.format_exc()
        success = False

    result_queue.put((success, output))


@linux_only
@skip_under_cuda_memcheck("May hang CUDA memcheck")
@skip_on_cudasim("Simulator does not require or implement MVC")
@skip_if_mvc_libraries_unavailable
class TestMinorVersionCompatibility(CUDATestCase):
    def test_mvc(self):
        # Run test with Minor Version Compatibility enabled in a child process
        ctx = mp.get_context("spawn")
        result_queue = ctx.Queue()
        proc = ctx.Process(target=child_test_wrapper, args=(result_queue,))
        proc.start()
        proc.join()
        success, output = result_queue.get()

        # Ensure the child process ran to completion before checking its output
        if not success:
            self.fail(output)


if __name__ == "__main__":
    unittest.main()



"""
Test cases adapted from numba/tests/test_enums.py
"""

import numpy as np

from numba import int16, int32
from numba import cuda, vectorize, njit
from numba.core import types
from numba.cuda.testing import unittest, CUDATestCase, skip_on_cudasim
from numba.tests.enum_usecases import (
    Color,
    Shape,
    Planet,
    RequestError,
    IntEnumWithNegatives,
)


class EnumTest(CUDATestCase):
    pairs = [
        (Color.red, Color.red),
        (Color.red, Color.green),
        (Planet.EARTH, Planet.EARTH),
        (Planet.VENUS, Planet.MARS),
        (Shape.circle, IntEnumWithNegatives.two),  # IntEnum, same value
    ]

    def test_compare(self):
        def f(a, b, out):
            out[0] = a == b
            out[1] = a != b
            out[2] = a is b
            out[3] = a is not b

        cuda_f = cuda.jit(f)
        for a, b in self.pairs:
            got = np.zeros((4,), dtype=np.bool_)
            expected = got.copy()
            cuda_f[1, 1](a, b, got)
            f(a, b, expected)
            self.assertPreciseEqual(expected, got)

    def test_getattr_getitem(self):
        def f(out):
            # Lookup of an enum member on its class
            out[0] = Color.red == Color.green
            out[1] = Color["red"] == Color["green"]

        cuda_f = cuda.jit(f)
        got = np.zeros((2,), dtype=np.bool_)
        expected = got.copy()
        cuda_f[1, 1](got)
        f(expected)
        self.assertPreciseEqual(expected, got)

    def test_return_from_device_func(self):
        @njit
        def inner(pred):
            return Color.red if pred else Color.green

        def f(pred, out):
            out[0] = inner(pred) == Color.red
            out[1] = inner(not pred) == Color.green

        cuda_f = cuda.jit(f)
        got = np.zeros((2,), dtype=np.bool_)
        expected = got.copy()
        f(True, expected)
        cuda_f[1, 1](True, got)
        self.assertPreciseEqual(expected, got)

    def test_int_coerce(self):
        def f(x, out):
            # Implicit coercion of intenums to ints
            if x > RequestError.internal_error:
                out[0] = x - RequestError.not_found
            else:
                out[0] = x + Shape.circle

        cuda_f = cuda.jit(f)
        for x in [300, 450, 550]:
            got = np.zeros((1,), dtype=np.int32)
            expected = got.copy()
            cuda_f[1, 1](x, got)
            f(x, expected)
            self.assertPreciseEqual(expected, got)

    def test_int_cast(self):
        def f(x, out):
            # Explicit coercion of intenums to ints
            if x > int16(RequestError.internal_error):
                out[0] = x - int32(RequestError.not_found)
            else:
                out[0] = x + int16(Shape.circle)

        cuda_f = cuda.jit(f)
        for x in [300, 450, 550]:
            got = np.zeros((1,), dtype=np.int32)
            expected = got.copy()
            cuda_f[1, 1](x, got)
            f(x, expected)
            self.assertEqual(expected, got)

    @skip_on_cudasim("ufuncs are unsupported on simulator.")
    def test_vectorize(self):
        def f(x):
            if x != RequestError.not_found:
                return RequestError["internal_error"]
            else:
                return RequestError.dummy

        cuda_func = vectorize("int64(int64)", target="cuda")(f)
        arr = np.array([2, 404, 500, 404], dtype=np.int64)
        expected = np.array([f(x) for x in arr], dtype=np.int64)
        got = cuda_func(arr)
        self.assertPreciseEqual(expected, got)

    def test_int_enum_no_conversion(self):
        # Ported from Numba PR #10047: "Fix IntEnumMember.can_convert_to() when
        # no conversions found", https://github.com/numba/numba/pull/10047.

        # The original test is intended to ensures that
        # IntEnumMember.can_convert_to() handles the case when the typing
        # context's can_convert() method returns None to signal no possible
        # conversion. In Numba-CUDA, we had to patch the CUDA target context to
        # work around this issue, because we cannot guarantee that the
        # IntEnumMember method can be patched before instances are created.
        ctx = cuda.descriptor.cuda_target.typing_context

        int_enum_type = types.IntEnumMember(Shape, types.int64)
        # Conversion of an int enum member to a 1D array would be invalid
        invalid_toty = types.int64[::1]
        self.assertIsNone(ctx.can_convert(int_enum_type, invalid_toty))


if __name__ == "__main__":
    unittest.main()



import threading
from numba import cuda
from numba.cuda.cudadrv.driver import driver
from numba.cuda.testing import unittest, ContextResettingTestCase
from queue import Queue


class TestResetDevice(ContextResettingTestCase):
    def test_reset_device(self):
        def newthread(exception_queue):
            try:
                devices = range(driver.get_device_count())
                for _ in range(2):
                    for d in devices:
                        cuda.select_device(d)
                        cuda.close()
            except Exception as e:
                exception_queue.put(e)

        # Do test on a separate thread so that we don't affect
        # the current context in the main thread.

        exception_queue = Queue()
        t = threading.Thread(target=newthread, args=(exception_queue,))
        t.start()
        t.join()

        exceptions = []
        while not exception_queue.empty():
            exceptions.append(exception_queue.get())
        self.assertEqual(exceptions, [])


if __name__ == "__main__":
    unittest.main()



import os
import platform
import shutil

from numba.tests.support import SerialMixin
from numba.cuda.cuda_paths import get_conda_ctk
from numba.cuda.cudadrv import driver, devices, libs
from numba.core import config
from numba.tests.support import TestCase
from pathlib import Path
import unittest

numba_cuda_dir = Path(__file__).parent
test_data_dir = numba_cuda_dir / "tests" / "data"


class CUDATestCase(SerialMixin, TestCase):
    """
    For tests that use a CUDA device. Test methods in a CUDATestCase must not
    be run out of module order, because the ContextResettingTestCase may reset
    the context and destroy resources used by a normal CUDATestCase if any of
    its tests are run between tests from a CUDATestCase.
    """

    def setUp(self):
        self._low_occupancy_warnings = config.CUDA_LOW_OCCUPANCY_WARNINGS
        self._warn_on_implicit_copy = config.CUDA_WARN_ON_IMPLICIT_COPY

        # Disable warnings about low gpu utilization in the test suite
        config.CUDA_LOW_OCCUPANCY_WARNINGS = 0
        # Disable warnings about host arrays in the test suite
        config.CUDA_WARN_ON_IMPLICIT_COPY = 0

    def tearDown(self):
        config.CUDA_LOW_OCCUPANCY_WARNINGS = self._low_occupancy_warnings
        config.CUDA_WARN_ON_IMPLICIT_COPY = self._warn_on_implicit_copy

    def skip_if_lto(self, reason):
        # Some linkers need the compute capability to be specified, so we
        # always specify it here.
        cc = devices.get_context().device.compute_capability
        linker = driver.Linker.new(cc=cc)
        if linker.lto:
            self.skipTest(reason)


class ContextResettingTestCase(CUDATestCase):
    """
    For tests where the context needs to be reset after each test. Typically
    these inspect or modify parts of the context that would usually be expected
    to be internal implementation details (such as the state of allocations and
    deallocations, etc.).
    """

    def tearDown(self):
        super().tearDown()
        from numba.cuda.cudadrv.devices import reset

        reset()


def ensure_supported_ccs_initialized():
    from numba.cuda import is_available as cuda_is_available
    from numba.cuda.cudadrv import nvvm

    if cuda_is_available():
        # Ensure that cudart.so is loaded and the list of supported compute
        # capabilities in the nvvm module is populated before a fork. This is
        # needed because some compilation tests don't require a CUDA context,
        # but do use NVVM, and it is required that libcudart.so should be
        # loaded before a fork (note that the requirement is not explicitly
        # documented).
        nvvm.get_supported_ccs()


def skip_on_cudasim(reason):
    """Skip this test if running on the CUDA simulator"""
    return unittest.skipIf(config.ENABLE_CUDASIM, reason)


def skip_unless_cudasim(reason):
    """Skip this test if running on CUDA hardware"""
    return unittest.skipUnless(config.ENABLE_CUDASIM, reason)


def skip_unless_conda_cudatoolkit(reason):
    """Skip test if the CUDA toolkit was not installed by Conda"""
    return unittest.skipUnless(get_conda_ctk() is not None, reason)


def skip_if_external_memmgr(reason):
    """Skip test if an EMM Plugin is in use"""
    return unittest.skipIf(config.CUDA_MEMORY_MANAGER != "default", reason)


def skip_under_cuda_memcheck(reason):
    return unittest.skipIf(os.environ.get("CUDA_MEMCHECK") is not None, reason)


def skip_without_nvdisasm(reason):
    nvdisasm_path = shutil.which("nvdisasm")
    return unittest.skipIf(nvdisasm_path is None, reason)


def skip_with_nvdisasm(reason):
    nvdisasm_path = shutil.which("nvdisasm")
    return unittest.skipIf(nvdisasm_path is not None, reason)


def skip_on_arm(reason):
    cpu = platform.processor()
    is_arm = cpu.startswith("arm") or cpu.startswith("aarch")
    return unittest.skipIf(is_arm, reason)


def skip_if_cuda_includes_missing(fn):
    # Skip when cuda.h is not available - generally this should indicate
    # whether the CUDA includes are available or not
    cuda_include_path = libs.get_cuda_include_dir()
    cuda_h = os.path.join(cuda_include_path, "cuda.h")
    cuda_h_file = os.path.exists(cuda_h) and os.path.isfile(cuda_h)
    reason = "CUDA include dir not available on this system"
    return unittest.skipUnless(cuda_h_file, reason)(fn)


def skip_if_curand_kernel_missing(fn):
    cuda_include_path = libs.get_cuda_include_dir()
    curand_kernel_h = os.path.join(cuda_include_path, "curand_kernel.h")
    curand_kernel_h_file = os.path.exists(curand_kernel_h) and os.path.isfile(
        curand_kernel_h
    )
    reason = "curand_kernel.h not available on this system"
    return unittest.skipUnless(curand_kernel_h_file, reason)(fn)


def skip_if_mvc_enabled(reason):
    """Skip a test if Minor Version Compatibility is enabled"""
    return unittest.skipIf(
        config.CUDA_ENABLE_MINOR_VERSION_COMPATIBILITY, reason
    )


def skip_if_mvc_libraries_unavailable(fn):
    libs_available = False
    try:
        import cubinlinker  # noqa: F401
        import ptxcompiler  # noqa: F401

        libs_available = True
    except ImportError:
        pass

    return unittest.skipUnless(
        libs_available, "Requires cubinlinker and ptxcompiler"
    )(fn)


def cc_X_or_above(major, minor):
    if not config.ENABLE_CUDASIM:
        cc = devices.get_context().device.compute_capability
        return cc >= (major, minor)
    else:
        return True


def skip_unless_cc_50(fn):
    return unittest.skipUnless(cc_X_or_above(5, 0), "requires cc >= 5.0")(fn)


def skip_unless_cc_53(fn):
    return unittest.skipUnless(cc_X_or_above(5, 3), "requires cc >= 5.3")(fn)


def skip_unless_cc_60(fn):
    return unittest.skipUnless(cc_X_or_above(6, 0), "requires cc >= 6.0")(fn)


def skip_unless_cc_75(fn):
    return unittest.skipUnless(cc_X_or_above(7, 5), "requires cc >= 7.5")(fn)


def xfail_unless_cudasim(fn):
    if config.ENABLE_CUDASIM:
        return fn
    else:
        return unittest.expectedFailure(fn)


def skip_with_cuda_python(reason):
    return unittest.skipIf(driver.USE_NV_BINDING, reason)


def cudadevrt_missing():
    if config.ENABLE_CUDASIM:
        return False
    try:
        path = libs.get_cudalib("cudadevrt", static=True)
        libs.check_static_lib(path)
    except FileNotFoundError:
        return True
    return False


def skip_if_cudadevrt_missing(fn):
    return unittest.skipIf(cudadevrt_missing(), "cudadevrt missing")(fn)


class ForeignArray(object):
    """
    Class for emulating an array coming from another library through the CUDA
    Array interface. This just hides a DeviceNDArray so that it doesn't look
    like a DeviceNDArray.
    """

    def __init__(self, arr):
        self._arr = arr
        self.__cuda_array_interface__ = arr.__cuda_array_interface__



from numba import types
from numba.cuda.stubs import _vector_type_stubs


class SimulatedVectorType:
    attributes = ["x", "y", "z", "w"]

    def __init__(self, *args):
        args_flattened = []
        for arg in args:
            if isinstance(arg, SimulatedVectorType):
                args_flattened += arg.as_list()
            else:
                args_flattened.append(arg)
        self._attrs = self.attributes[: len(args_flattened)]
        if not self.num_elements == len(args_flattened):
            raise TypeError(
                f"{self.name} expects {self.num_elements}"
                f" elements, got {len(args_flattened)}"
            )

        for arg, attr in zip(args_flattened, self._attrs):
            setattr(self, attr, arg)

    @property
    def name(self):
        raise NotImplementedError()

    @property
    def num_elements(self):
        raise NotImplementedError()

    def as_list(self):
        return [getattr(self, attr) for attr in self._attrs]


def make_simulated_vector_type(num_elements, name):
    obj = type(
        name,
        (SimulatedVectorType,),
        {
            "num_elements": num_elements,
            "base_type": types.float32,
            "name": name,
        },
    )
    obj.user_facing_object = obj
    return obj


def _initialize():
    _simulated_vector_types = {}
    for stub in _vector_type_stubs:
        num_elements = int(stub.__name__[-1])
        _simulated_vector_types[stub.__name__] = make_simulated_vector_type(
            num_elements, stub.__name__
        )
        _simulated_vector_types[stub.__name__].aliases = stub.aliases
    return _simulated_vector_types


vector_types = _initialize()



import functools
import numpy as np
import unittest

from numba import config, cuda, types
from numba.tests.support import TestCase
from numba.tests.test_ufuncs import BasicUFuncTest


def _make_ufunc_usecase(ufunc):
    ldict = {}
    arg_str = ",".join(["a{0}".format(i) for i in range(ufunc.nargs)])
    func_str = f"def fn({arg_str}):\n    np.{ufunc.__name__}({arg_str})"
    exec(func_str, globals(), ldict)
    fn = ldict["fn"]
    fn.__name__ = "{0}_usecase".format(ufunc.__name__)
    return fn


# This test would also be a CUDATestCase, but to avoid a confusing and
# potentially dangerous inheritance diamond with setUp methods that modify
# global state, we implement the necessary parts of CUDATestCase within this
# class instead. These are:
#
# - Disable parallel testing with _numba_parallel_test_.
# - Disabling CUDA performance warnings for the duration of tests.
class TestUFuncs(BasicUFuncTest, TestCase):
    _numba_parallel_test_ = False

    def setUp(self):
        BasicUFuncTest.setUp(self)

        # The basic ufunc test does not set up complex inputs, so we'll add
        # some here for testing with CUDA.
        self.inputs.extend(
            [
                (np.complex64(-0.5 - 0.5j), types.complex64),
                (np.complex64(0.0), types.complex64),
                (np.complex64(0.5 + 0.5j), types.complex64),
                (np.complex128(-0.5 - 0.5j), types.complex128),
                (np.complex128(0.0), types.complex128),
                (np.complex128(0.5 + 0.5j), types.complex128),
                (
                    np.array([-0.5 - 0.5j, 0.0, 0.5 + 0.5j], dtype="c8"),
                    types.Array(types.complex64, 1, "C"),
                ),
                (
                    np.array([-0.5 - 0.5j, 0.0, 0.5 + 0.5j], dtype="c16"),
                    types.Array(types.complex128, 1, "C"),
                ),
            ]
        )

        # Test with multiple dimensions
        self.inputs.extend(
            [
                # Basic 2D and 3D arrays
                (
                    np.linspace(0, 1).reshape((5, -1)),
                    types.Array(types.float64, 2, "C"),
                ),
                (
                    np.linspace(0, 1).reshape((2, 5, -1)),
                    types.Array(types.float64, 3, "C"),
                ),
                # Complex data (i.e. interleaved)
                (
                    np.linspace(0, 1 + 1j).reshape(5, -1),
                    types.Array(types.complex128, 2, "C"),
                ),
                # F-ordered
                (
                    np.asfortranarray(np.linspace(0, 1).reshape((5, -1))),
                    types.Array(types.float64, 2, "F"),
                ),
            ]
        )

        # Add tests for other integer types
        self.inputs.extend(
            [
                (np.uint8(0), types.uint8),
                (np.uint8(1), types.uint8),
                (np.int8(-1), types.int8),
                (np.int8(0), types.int8),
                (np.uint16(0), types.uint16),
                (np.uint16(1), types.uint16),
                (np.int16(-1), types.int16),
                (np.int16(0), types.int16),
                (np.ulonglong(0), types.ulonglong),
                (np.ulonglong(1), types.ulonglong),
                (np.longlong(-1), types.longlong),
                (np.longlong(0), types.longlong),
                (
                    np.array([0, 1], dtype=np.ulonglong),
                    types.Array(types.ulonglong, 1, "C"),
                ),
                (
                    np.array([0, 1], dtype=np.longlong),
                    types.Array(types.longlong, 1, "C"),
                ),
            ]
        )

        self._low_occupancy_warnings = config.CUDA_LOW_OCCUPANCY_WARNINGS
        self._warn_on_implicit_copy = config.CUDA_WARN_ON_IMPLICIT_COPY

        # Disable warnings about low gpu utilization in the test suite
        config.CUDA_LOW_OCCUPANCY_WARNINGS = 0
        # Disable warnings about host arrays in the test suite
        config.CUDA_WARN_ON_IMPLICIT_COPY = 0

    def tearDown(self):
        # Restore original warning settings
        config.CUDA_LOW_OCCUPANCY_WARNINGS = self._low_occupancy_warnings
        config.CUDA_WARN_ON_IMPLICIT_COPY = self._warn_on_implicit_copy

    def _make_ufunc_usecase(self, ufunc):
        return _make_ufunc_usecase(ufunc)

    @functools.lru_cache(maxsize=None)
    def _compile(self, pyfunc, args):
        # We return an already-configured kernel so that basic_ufunc_test can
        # call it just like it does for a CPU function
        return cuda.jit(args)(pyfunc)[1, 1]

    def basic_int_ufunc_test(self, name=None):
        skip_inputs = [
            types.float32,
            types.float64,
            types.Array(types.float32, 1, "C"),
            types.Array(types.float32, 2, "C"),
            types.Array(types.float64, 1, "C"),
            types.Array(types.float64, 2, "C"),
            types.Array(types.float64, 3, "C"),
            types.Array(types.float64, 2, "F"),
            types.complex64,
            types.complex128,
            types.Array(types.complex64, 1, "C"),
            types.Array(types.complex64, 2, "C"),
            types.Array(types.complex128, 1, "C"),
            types.Array(types.complex128, 2, "C"),
        ]
        self.basic_ufunc_test(name, skip_inputs=skip_inputs)

    ############################################################################
    # Trigonometric Functions

    def test_sin_ufunc(self):
        self.basic_ufunc_test(np.sin, kinds="cf")

    def test_cos_ufunc(self):
        self.basic_ufunc_test(np.cos, kinds="cf")

    def test_tan_ufunc(self):
        self.basic_ufunc_test(np.tan, kinds="cf")

    def test_arcsin_ufunc(self):
        self.basic_ufunc_test(np.arcsin, kinds="cf")

    def test_arccos_ufunc(self):
        self.basic_ufunc_test(np.arccos, kinds="cf")

    def test_arctan_ufunc(self):
        self.basic_ufunc_test(np.arctan, kinds="cf")

    def test_arctan2_ufunc(self):
        self.basic_ufunc_test(np.arctan2, kinds="f")

    def test_hypot_ufunc(self):
        self.basic_ufunc_test(np.hypot, kinds="f")

    def test_sinh_ufunc(self):
        self.basic_ufunc_test(np.sinh, kinds="cf")

    def test_cosh_ufunc(self):
        self.basic_ufunc_test(np.cosh, kinds="cf")

    def test_tanh_ufunc(self):
        self.basic_ufunc_test(np.tanh, kinds="cf")

    def test_arcsinh_ufunc(self):
        self.basic_ufunc_test(np.arcsinh, kinds="cf")

    def test_arccosh_ufunc(self):
        self.basic_ufunc_test(np.arccosh, kinds="cf")

    def test_arctanh_ufunc(self):
        # arctanh is only valid is only finite in the range ]-1, 1[
        # This means that for any of the integer types it will produce
        # conversion from infinity/-infinity to integer. That's undefined
        # behavior in C, so the results may vary from implementation to
        # implementation. This means that the result from the compiler
        # used to compile NumPy may differ from the result generated by
        # llvm. Skipping the integer types in this test avoids failed
        # tests because of this.
        to_skip = [
            types.Array(types.uint32, 1, "C"),
            types.uint32,
            types.Array(types.int32, 1, "C"),
            types.int32,
            types.Array(types.uint64, 1, "C"),
            types.uint64,
            types.Array(types.int64, 1, "C"),
            types.int64,
        ]

        self.basic_ufunc_test(np.arctanh, skip_inputs=to_skip, kinds="cf")

    def test_deg2rad_ufunc(self):
        self.basic_ufunc_test(np.deg2rad, kinds="f")

    def test_rad2deg_ufunc(self):
        self.basic_ufunc_test(np.rad2deg, kinds="f")

    def test_degrees_ufunc(self):
        self.basic_ufunc_test(np.degrees, kinds="f")

    def test_radians_ufunc(self):
        self.basic_ufunc_test(np.radians, kinds="f")

    ############################################################################
    # Comparison functions
    def test_greater_ufunc(self):
        self.signed_unsigned_cmp_test(np.greater)

    def test_greater_equal_ufunc(self):
        self.signed_unsigned_cmp_test(np.greater_equal)

    def test_less_ufunc(self):
        self.signed_unsigned_cmp_test(np.less)

    def test_less_equal_ufunc(self):
        self.signed_unsigned_cmp_test(np.less_equal)

    def test_not_equal_ufunc(self):
        self.signed_unsigned_cmp_test(np.not_equal)

    def test_equal_ufunc(self):
        self.signed_unsigned_cmp_test(np.equal)

    def test_logical_and_ufunc(self):
        self.basic_ufunc_test(np.logical_and)

    def test_logical_or_ufunc(self):
        self.basic_ufunc_test(np.logical_or)

    def test_logical_xor_ufunc(self):
        self.basic_ufunc_test(np.logical_xor)

    def test_logical_not_ufunc(self):
        self.basic_ufunc_test(np.logical_not)

    def test_maximum_ufunc(self):
        self.basic_ufunc_test(np.maximum)

    def test_minimum_ufunc(self):
        self.basic_ufunc_test(np.minimum)

    def test_fmax_ufunc(self):
        self.basic_ufunc_test(np.fmax)

    def test_fmin_ufunc(self):
        self.basic_ufunc_test(np.fmin)

    def test_bitwise_and_ufunc(self):
        self.basic_int_ufunc_test(np.bitwise_and)

    def test_bitwise_or_ufunc(self):
        self.basic_int_ufunc_test(np.bitwise_or)

    def test_bitwise_xor_ufunc(self):
        self.basic_int_ufunc_test(np.bitwise_xor)

    def test_invert_ufunc(self):
        self.basic_int_ufunc_test(np.invert)

    def test_bitwise_not_ufunc(self):
        self.basic_int_ufunc_test(np.bitwise_not)

    # Note: there is no entry for np.left_shift and np.right_shift
    # because their implementations in NumPy have undefined behavior
    # when the second argument is a negative. See the comment in
    # numba/tests/test_ufuncs.py for more details.

    ############################################################################
    # Mathematical Functions

    def test_log_ufunc(self):
        self.basic_ufunc_test(np.log, kinds="cf")

    def test_log2_ufunc(self):
        self.basic_ufunc_test(np.log2, kinds="cf")

    def test_log10_ufunc(self):
        self.basic_ufunc_test(np.log10, kinds="cf")


if __name__ == "__main__":
    unittest.main()



import functools

from llvmlite import ir

from numba.core.datamodel.registry import DataModelManager, register
from numba.core.extending import models
from numba.core import types
from numba.cuda.types import Dim3, GridGroup, CUDADispatcher


cuda_data_manager = DataModelManager()

register_model = functools.partial(register, cuda_data_manager)


@register_model(Dim3)
class Dim3Model(models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [("x", types.int32), ("y", types.int32), ("z", types.int32)]
        super().__init__(dmm, fe_type, members)


@register_model(GridGroup)
class GridGroupModel(models.PrimitiveModel):
    def __init__(self, dmm, fe_type):
        be_type = ir.IntType(64)
        super().__init__(dmm, fe_type, be_type)


@register_model(types.Float)
class FloatModel(models.PrimitiveModel):
    def __init__(self, dmm, fe_type):
        if fe_type == types.float16:
            be_type = ir.IntType(16)
        elif fe_type == types.float32:
            be_type = ir.FloatType()
        elif fe_type == types.float64:
            be_type = ir.DoubleType()
        else:
            raise NotImplementedError(fe_type)
        super(FloatModel, self).__init__(dmm, fe_type, be_type)


register_model(CUDADispatcher)(models.OpaqueModel)
